// ****************************************************************************
//             Copyright (C) Oasys Design Systems, Inc. 2004 - 2008
//                             All rights reserved.
// ****************************************************************************
#include "LLoad.h"
#include "LibLoadMsg.h"
#include "attr_enum.h"
#include "NCell"
#include "NMVCell"
#include "UParam"
#include "TGenOpCnd"

namespace UParam {
  extern Variable<double> areaUnit;
  extern Variable<bool> markPadAsMacro;
}

// Gated format has to be _ separated keywords like ff_rise_precontrol_obs
int
LLoad::processClkGateAttr(const char* attr, TEdge& edge, NClkGtRegType& seqType,
			  NClkGtTestType& test)
{
  char buf[50];
  strcpy(buf, attr);
  strcat(buf, "_");
  char* c = buf;
  char* n = strchr(c, '_');
  if(n == NULL) {
    // impossible since we just appended '_' but...
    return 0;
  }
  *n = 0;
  UString str(c);
  switch (str.keyword()) {
  case K_none:
    seqType = NCGRT_none;
    break;
  case K_latch:
    seqType = NCGRT_latch;
    break;
  case K_ff:
    seqType = NCGRT_ff;
    break;
  default:
    // error
    return 0;
  }

  c = n+1;
  if (*c == 0) {
    // error
    return 0;
  }
  n = strchr(c, '_');
  if(n == NULL) {
    // error
    return 0;
  }
  *n = 0;
  str = c;
  switch (str.keyword()) {
  case K_posedge:
    edge = TEdge::rise;
    break;
  case K_negedge:
    edge = TEdge::fall;
    break;
  default:
    // error
    return 0;
  }

  test = NCGTT_none;
  c = n+1;
  if (*c == 0) {
    return 1;
  }
  n = strchr(c, '_');
  if(n == NULL) {
    // error
    return 0;
  }
  *n = 0;
  str = c;
  switch (str.keyword()) {
  case K_control:
  case K_precontrol:
    test = NCGTT_preControl;
    break;
  case K_postcontrol:
    test = NCGTT_postControl;
    break;
  default:
    // error
    return 0;
  }

  c = n+1;
  if (*c == 0) {
    return 1;
  }
  n = strchr(c, '_');
  if(n == NULL) {
    // error
    return 0;
  }
  *n = 0;
  str = c;
  switch (str.keyword()) {
  case K_obs:
    test = NCGTT_preControl ? NCGTT_preControlObs : NCGTT_postControlObs;
    break;
  default:
    // error
    return 0;
  }

  c = n+1;
  if (*c != 0) {
    // error
    return 0;
  }
  return 1;
}

void
LLoad::preProcessCellAttr(const LibGroup* lcell, bool& isMvCell)
{
  LibAttrIter aiter;
  LibAttr* attr;
  isMvCell   = false;

  aiter = lcell->attrIter();
  while ((attr = aiter++) && !isMvCell) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_always_on                    : // boolean
    case LIBERTY_ATTRENUM_is_isolation_cell            : // boolean;
    case LIBERTY_ATTRENUM_is_level_shifter             : // boolean;
    case LIBERTY_ATTRENUM_power_gating_cell            : // string;
    case LIBERTY_ATTRENUM_retention_cell               : // string;
      // cannot create MV cell for all dont_use cell, because
      // some standard cells are really being marked dont_use in libraries.
      //case LIBERTY_ATTRENUM_dont_use                     : // boolean
      //case LIBERTY_ATTRENUM_dont_touch                   : // boolean
      {
	isMvCell = true;
      }
      break;
    default                                            : 
      break;
    };
  }
}

void
LLoad::processCellAttr(const LibGroup* lcell, NCell cell,
		       LLoadRailList& railList)
{
  LibAttrIter aiter;
  LibAttr* attr;
  UString funcClass;
  UString vthGroup;
  TLib* tlib = d_lib.tlib();
  const TLibScaleFactors* sf;

  aiter = lcell->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_area                         : // float( > 0.0 ) := 0.0;
      uassert (round(attr->value()->double_val() / UParam::areaUnit()) < ULONG_LONG_MAX); // make sure there's no overflow
      cell.area((ullong_t)round(attr->value()->double_val() / UParam::areaUnit()));
      break;
    case LIBERTY_ATTRENUM_dont_touch                   : // boolean;
      cell.setDontTouch(attr->value()->bool_val());
      break;
    case LIBERTY_ATTRENUM_dont_use                     : // boolean;
      cell.setDontUse(attr->value()->bool_val());
      break;
    case LIBERTY_ATTRENUM_dont_remap                   : // boolean;
      cell.setDontRemap(attr->value()->bool_val());
      break;
    case LIBERTY_ATTRENUM_scaling_factors              : // string;
      if ((sf = tlib->findScaleFactors(attr->value()->string_val()))) {
	cell.scaleFactors(sf, MVCRNR_INITIAL);
      } else {
	TLIBMSG(scalingFactorsNotFound, attr->value()->string_val());
      }
      break;
    case LIBERTY_ATTRENUM_user_function_class          : // string;
      funcClass = attr->value()->string_val();
      cell.userFuncClass(funcClass);
      break;
    case LIBERTY_ATTRENUM_cell_leakage_power           : // float( > 0.0 );
      cell.setLeakagePower(tlib->convertLeakagePower(attr->value()->double_val()),
			   MVCRNR_INITIAL);
      break;
    case LIBERTY_ATTRENUM_clock_gating_integrated_cell : // string( "latch_posedge", "latch_posedge_precontrol", "latch_posedge_postcontrol", "latch_posedge_precontrol_obs", "latch_posedge_postcontrol_obs", "latch_negedge", "latch_negedge_precontrol", "latch_negedge_postcontrol", "latch_negedge_precontrol_obs", "latch_negedge_postcontrol_obs", "ff_posedge", "ff_posedge_precontrol", "ff_posedge_postcontrol", "ff_posedge_precontrol_obs", "ff_posedge_postcontrol_obs", "ff_negedge", "ff_negedge_precontrol", "ff_negedge_postcontrol", "ff_negedge_precontrol_obs", "ff_negedge_postcontrol_obs", "none_posedge", "none_posedge_control", "none_posedge_control_obs", "none_negedge", "none_negedge_control", or "none_negedge_control_obs" );
      {
	NClkGtRegType  seqType = NCGRT_none;
	NClkGtTestType test    = NCGTT_none;
	TEdge          edge    = TEdge::rise;
	
	const char* clockGateType = attr->value()->string_val();
	if (processClkGateAttr(clockGateType, edge, seqType, test)) {
	  cell.setClockGate(edge == TEdge::rise ? false : true, seqType, test);
	} else {
	  UStrBuf cellName = cell.name();
	  LIBLOADMSG(invalidClkGateAttribute, (const char*)cellName, clockGateType);
	}
	break;
      }
    case LIBERTY_ATTRENUM_pad_cell                     : // boolean;
      if( attr->value()->bool_val() ) {
	cell.setPad();
#ifdef RT_FPGA
	if(UParam::markPadAsMacro()) {
	  cell.sequentialType(NST_macro);
	}
#else
	cell.sequentialType(NST_macro);
#endif
      }
      break;
    case LIBERTY_ATTRENUM_is_clock_gating_cell         : // boolean;
      if (attr->value()->bool_val()) {
	cell.setClockGate(false /*should be determined from function*/, NCGRT_none, NCGTT_none);
      }
      break;
    case LIBERTY_ATTRENUM_always_on                    : // boolean;
      if( attr->value()->bool_val() ) {
	NMVCell mvCell = cell.getMVCell();
	if (mvCell.mvCellType() == NMVT_standard) {
	  // default to backup header first, and will be amended
	  // in createPGPins() if it is a footer always_on.
	  mvCell.setPowerType(NPT_backupHeader);
	}
      }
      break;
    case LIBERTY_ATTRENUM_is_isolation_cell            : // boolean;
      if( attr->value()->bool_val() ) {
	NMVCell mvCell = cell.getMVCell();
	if (mvCell.mvCellType() == NMVT_lshifter) {
	  mvCell.setMvCellType(NMVT_elshifter);
	} else {
	  mvCell.setMvCellType(NMVT_isolation);
	}
      }
      break;
    case LIBERTY_ATTRENUM_is_level_shifter             : // boolean;
      if( attr->value()->bool_val() ) {
	NMVCell mvCell = cell.getMVCell();
	if (mvCell.mvCellType() == NMVT_isolation) {
	  mvCell.setMvCellType(NMVT_elshifter);
	} else {
	  mvCell.setMvCellType(NMVT_lshifter);
	}
      }
      break;
    case LIBERTY_ATTRENUM_power_gating_cell            : // string;
    case LIBERTY_ATTRENUM_retention_cell               : // string;
      {
	const char* retClass = attr->value()->string_val();
	NMVCell mvCell = cell.getMVCell();
	mvCell.setMvCellType(NMVT_retention);
	mvCell.setRetentionClass(retClass);
      }
      break;
    case LIBERTY_ATTRENUM_input_voltage_range          : // (float, float)
    case LIBERTY_ATTRENUM_output_voltage_range         : // (float, float)
      {
	NMVCell mvCell = cell.getMVCell();
	if (mvCell) {
	  //uassert(mvCell.mvCellType() & NMVT_lshifter);
	  // is_level_shifter may come after input_voltage_range, etc.
	  // so, should set it here first
	  if (mvCell.mvCellType() == NMVT_isolation) {
	    mvCell.setMvCellType(NMVT_elshifter);
	  } else {
	    mvCell.setMvCellType(NMVT_lshifter);
	  }
	  LibValIter viter = attr->valueIter();
	  LibVal* v = viter++;
	  UFloat from = v->double_val();
	  v = viter++;
	  UFloat to   = v->double_val();
	  TValRange r;
	  r.addValue(from);
	  r.addValue(to);
	  if (attr->typeID() == LIBERTY_ATTRENUM_input_voltage_range) {
	    mvCell.setInputVoltageRange(r);
	  } else {
	    mvCell.setOutputVoltageRange(r);
	  }
	} else {
	  // ignore
	}
      }
      break;
    case LIBERTY_ATTRENUM_rail_connection              : // ( string, string );
      {
	LibValIter viter = attr->valueIter();
	LibVal* v = viter++;
	const char* railName   = v->string_val();
	v = viter++;
	const char* supplyName = v->string_val();
	LLoadRail rail(railName, supplyName);
	railList.append(rail);
      }
      break;
    case LIBERTY_ATTRENUM_threshold_voltage_group      : // string;
      {
	vthGroup = attr->value()->string_val();
	cell.setVthGroup(vthGroup);
      }
      break;
    case LIBERTY_ATTRENUM_timing_model_type            : // string;
    case LIBERTY_ATTRENUM_auxiliary_pad_cell           : // boolean;
    case LIBERTY_ATTRENUM_bus_naming_style             : // virtual_attribute;
    case LIBERTY_ATTRENUM_cell_footprint               : // string;
    case LIBERTY_ATTRENUM_contention_condition         : // string;
    case LIBERTY_ATTRENUM_dont_fault                   : // string( "sa0", "sa1", or "sa01" );
    case LIBERTY_ATTRENUM_geometry_print               : // string;
    case LIBERTY_ATTRENUM_handle_negative_constraint   : // boolean;
    case LIBERTY_ATTRENUM_interface_timing             : // boolean;
    case LIBERTY_ATTRENUM_map_only                     : // boolean;
    case LIBERTY_ATTRENUM_mpm_libname                  : // string;
    case LIBERTY_ATTRENUM_mpm_name                     : // string;
    case LIBERTY_ATTRENUM_observe_node                 : // string;
    case LIBERTY_ATTRENUM_pad_type                     : // enum( clock );
    case LIBERTY_ATTRENUM_preferred                    : // boolean;
    case LIBERTY_ATTRENUM_scan_group                   : // string;
    case LIBERTY_ATTRENUM_set_node                     : // string;
    case LIBERTY_ATTRENUM_single_bit_degenerate        : // string;
    case LIBERTY_ATTRENUM_use_for_size_only            : // boolean;
    case LIBERTY_ATTRENUM_vhdl_name                    : // string;
    case LIBERTY_ATTRENUM_pin_equal                    : // ( string );
    case LIBERTY_ATTRENUM_pin_opposite                 : // ( string, string );
    default                                            : 
      if (warnNoSupport()) {
	printf("Attribute %s (in cell %s) on line %d of library file %s"
	       " not supported.\n", (str_t)attr->name(), (str_t)cell.name(),
	       attr->lineno(), attr->filename());
      }
      break;
    };
  }
}

// This API effectively creates primary_power PG pins at LCell.
// Called from instantiaCellPins : rail_connection
void
LLoad::processRailConnAttrs(LCell &lcell, NCell cell, 
			    const LLoadRailList& railList)
{
  UAggIter pginPinIter = lcell.inPGPins().msbToLsb();
  UAggIter* pgiter     = &pginPinIter;
  TLibDir tdir         = TLD_input; // rail connection are assumed to be input
  TLib* tlib           = d_lib.tlib();

  for (LLoadRailIter li = railList.begin(); li ; li++) {
    LLoadRail rail = li.object();
    const UString& railName = rail.k_value();
    const UString& supplyName = rail.o_value();

    TLibPGPin tpin(tlib, railName, tdir);
    int idx = (*pgiter)++.current();
    lcell.pgpin(idx)->grab(tpin);
    UFloat volt = tlib->voltageMap(supplyName);
    tpin.ptr()->setVoltage(MVCRNR_INITIAL, volt);
    tpin.ptr()->setVoltageName(supplyName);

    // set the mainRail info based on default_power_rail info, if available
    if (supplyName == tlib->defaultPowerRail()) {
      tpin.ptr()->setMainRail(true);
    }
  }
}
