// ****************************************************************************
//             Copyright (C) Oasys Design Systems, Inc. 2004 - 2008
//                             All rights reserved.
// ****************************************************************************
#include "LLoad.h"
#include "LLoadFuncParse.h"
#include "group_enum.h"
#include "attr_enum.h"
#include "mymalloc.h"
#include "NCell"
#include "NMVCell"
#include "NPort"
#include "TLibPin"
#include "LibParam.h"
#include "LibLoadMsg.h"
#include "NDb"

UAggType*
LLoad::findType(const UString& nm, LCell& lcell)
{
  UAggType* t = lcell.findType(nm);
  if (t) {
    return t;
  }
  TLib* tlib = d_lib.tlib();
  t = tlib->findType(nm);
  uassert (t);
  return t;
}

void
LLoad::processSeqInfo(const LibGroup* seq, TLibInternalPinC* tpin, 
		      uint_t seqType, NCell cell, int outCount,
		      LLoadPinMap& pinMap)
{
  LibErrorT err;
  LibAttrIter aiter;
  LibAttr* attr;
  TLibSeqInfo* seqInfo = NULL;
  Cudd& bddMgr = d_lib.bddMgr();

  switch (seqType) {
  case LIBERTY_GROUPENUM_ff :
    if (seq->findAttr("clocked_on_also", err)) {
      // master-slave
      seqInfo = new IN_SAME_POOL_AS(tpin) TLibMSFFInfo(1);
      cell.sequentialType(NST_msff);
    } else {
      seqInfo = new IN_SAME_POOL_AS(tpin) TLibFFInfo(1);
      cell.sequentialType(NST_ff);
    }
    break;
  case LIBERTY_GROUPENUM_latch :
    seqInfo = new IN_SAME_POOL_AS(tpin) TLibLatchInfo(1);
    cell.sequentialType(NST_latch);
    break;
  }
  tpin->seqInfo(seqInfo);
  aiter = seq->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_data_in:
    case LIBERTY_ATTRENUM_next_state:
      seqInfo->nextState(parseFunc(attr->value()->string_val(),
				   bddMgr, pinMap, /*bundles=*/NULL, 
				   /*busses=*/NULL, /*busType=*/NULL,
				   /*grpIdx=*/-1, outCount));
      break;
    case LIBERTY_ATTRENUM_enable:
    case LIBERTY_ATTRENUM_clocked_on:
      seqInfo->clock(parseFunc(attr->value()->string_val(),
			       bddMgr, pinMap, /*bundles=*/NULL,
			       /*busses=*/NULL, /*busType=*/NULL,
			       /*grpIdx=*/-1, outCount));
      break;
    case LIBERTY_ATTRENUM_enable_also:
    case LIBERTY_ATTRENUM_clocked_on_also:
      ((TLibMSFFInfo*)seqInfo)->slaveClock(parseFunc(attr->value()->string_val(),
						     bddMgr, pinMap, /*bundles=*/NULL,
						     /*busses=*/NULL, /*busType=*/NULL,
						     /*grpIdx=*/-1, outCount));
      break;
    case LIBERTY_ATTRENUM_clear:
      seqInfo->clear(parseFunc(attr->value()->string_val(),
			       bddMgr, pinMap, /*bundles=*/NULL,
			       /*busses=*/NULL, /*busType=*/NULL,
			       /*grpIdx=*/-1, outCount));
      break;
    case LIBERTY_ATTRENUM_preset:
      seqInfo->preset(parseFunc(attr->value()->string_val(),
				bddMgr, pinMap, /*bundles=*/NULL,
				/*busses=*/NULL, /*busType=*/NULL,
				/*grpIdx=*/-1, outCount));
      break;
    case LIBERTY_ATTRENUM_clear_preset_var1:
      seqInfo->clearPresetVal1(attr->value()->string_val());
      break;
    case LIBERTY_ATTRENUM_clear_preset_var2:
      seqInfo->clearPresetVal2(attr->value()->string_val());
      break;
    default:
      break;
    }
  }
}

static NDb localNDb = NULL;

static pint_t
getFunctionPtr(const BDD& f)
{
  if (UParam::sortOutputPinsByIndex()) {
    pint_t idx = localNDb.libPinIdx(f.ptr());
    uassert(idx > 0);
    return idx;
  } else {
    return (pint_t)f.ptr();
  }
}

static pint_t
getNodeIndex(const BDD& f)
{
  return (pint_t)f.NodeReadIndex();
}

static int
compareCellDrivePins(TLibPin pinI, TLibPin pinJ, pint_t (*keyF)(const BDD&))
{
  pint_t i = keyF(pinI.function());
  pint_t j = keyF(pinJ.function());
  bool i_so = pinI.output()->isTestOutOnly();
  bool j_so = pinJ.output()->isTestOutOnly();
  if (!i_so) {
    if (j_so) {
      return -1;
    }
  } else if (!j_so) {
    return 1;
  }
  if (i < j) {
    return -1;
  } else if (i > j) {
    return 1;
  }
  return 0;
}

static void
qSortCellDrivePins(LCell& cell, URange r, pint_t (*keyF)(const BDD&))
{
  int high;
  int h = r.max();
  int l = r.min();
  if ((h - l) < 1) return;
  { // start a new body to prevent variables used only in this body
    // from going on to the stack in the recursion;
    TLibPin pivot = cell.pin(l);
    int low = l+1;
    high = h;
    while (low <= high) {
      while (low <= high) {
	if (compareCellDrivePins(cell.pin(low), pivot, keyF) >= 0) { 
	  break;
	} else {
	  low++;
	}
      }
      while (low <= high) {
	if (compareCellDrivePins(cell.pin(high), pivot, keyF) < 0) {
	  break;
	} else {
	  high--;
	}
      }
      if (low < high) {
	cell.swapPins(low, high);
	low++;
	high--;
      }
    }
  }
  if (high != l) {
    cell.swapPins(l, high);
  }
  if (high - 1 > l) {
    qSortCellDrivePins(cell, URange(l, high-1), keyF);
  }
  if (high + 1 < h) {
    qSortCellDrivePins(cell, URange(high+1, h), keyF);
  }
}

static void 
setCellPinIndices(NDb ndb, LCell& cell, const UAggSlice& slice)
{
  for (int i = slice.offset(); i < slice.offset() + slice.count(); i++) {
    BDD bdd = cell.pin(i).function();
    DdNode* dd = bdd.ptr();
    ndb.addLibPinIdx(dd);
  }
}


// canonicalize output/bidir pin order based on BDD ptr value
static void
canonicalizeDrivePinOrder(LCell& cell, pint_t (*keyF)(const BDD&))
{
  UAggSlice outPins   = cell.outPins();
  UAggSlice bidirPins = cell.bidirPins();
  int start, end;

  localNDb = cell.ncell().library(MVCRNR_INITIAL).database();
  
  setCellPinIndices(localNDb, cell, outPins);
  setCellPinIndices(localNDb, cell, bidirPins);
  
  start = outPins.offset();
  end   = start + outPins.count() - 1;
  if (end >= start) {
    qSortCellDrivePins(cell, URange(start, end), keyF);
  }
  start = bidirPins.offset();
  end   = start + bidirPins.count() - 1;
  if (end >= start) {
    qSortCellDrivePins(cell, URange(start, end), keyF);
  }

  localNDb = NULL;
}

// No point in trying to canonicalize input pin order
// if there are output/bidir pins with no function since
// such cells cannot be classified into any family,
// except by user func, in which case even the original
// pin order from the library is sufficient

static bool
pointless2Canonicalize(LCell& cell)
{
  UAggIter iter;

  for (iter = cell.driverPins().msbToLsb(); iter; iter++) {
    if (!cell.pin(iter).function().ptr()) {
      return true;
    }
  }
  for (iter = cell.internalPins().msbToLsb(); iter; iter++) {
    if (!cell.pin(iter).seqInfo()) {
      return true;
    }
  }
  return false;
}

void
updatePriority(BDD& f, BDD var, int priority[], int from, int to)
{
  BDD   fx    = f.Restrict(var);
  BDD   fxbar = f.Restrict(!var);
  int*  fxArr = f.manager()->supportArray(0);
  int*  fxbarArr = f.manager()->supportArray(1);
  int   i;
  
  // prioritize variables in the support of fxbar
  fxbar.SupportIndex(fxbarArr);
  fx.SupportIndex(fxArr);
  for (int idx = from; idx < to; idx++) {
    i = idx - from;
    if (priority[i] > -1) {
      if (fxbarArr[idx]) {
	priority[i] *= 3;
	if (!fxArr[idx]) {
	  priority[i] += 2;
	}
      } else if (fxArr[idx]) {
	priority[i] = priority[i] * 3 + 1;
      }
    }
  }
}

struct VarInfo
{
public:
  VarInfo() : name(UString::Null), nextStateType(UString::Null), idx(INT_MAX),
	      bdSupport(-1), bdMinterm(-1), fxbarMinterm(-1), fxMinterm(-1),
	      priority(-1) {}
  VarInfo(const UString& nm, const UString& nsType, int i, int s, double m1,
	  double m2, double m3, int p) :
    name(nm), nextStateType(nsType), idx(i), bdSupport(s), bdMinterm(m1),
    fxbarMinterm(m2), fxMinterm(m3),
    priority(p) {}

  bool precedes(const VarInfo& nvar) const; 
  
  UString name;
  UString nextStateType;
  int    idx;
  int    bdSupport;    // support of fx ^ fxbar (boolean difference)
  double bdMinterm;    // number of minterms in boolean difference
  double fxbarMinterm; // number of minterms in fxbar
  double fxMinterm;    // number of minterms in fx
  int    priority;     // priority order of variable based on
                       // other already ordered variables
};

inline bool
VarInfo::precedes(const VarInfo& nvar) const
{
  if (bdSupport > nvar.bdSupport) {
    return true;
  } else if (bdSupport < nvar.bdSupport) {
    return false;
  } else if (bdMinterm > nvar.bdMinterm) {
    return true;
  } else if (bdMinterm < nvar.bdMinterm) {
    return false;
  } else if (priority > nvar.priority) {
    return true;
  } else if (priority < nvar.priority) {
    return false;
  } else if (fxbarMinterm > nvar.fxbarMinterm) {
    return true;
  } else if (fxbarMinterm < nvar.fxbarMinterm) {
    return false;
  } else if (fxMinterm > nvar.fxMinterm) {
    return true;
  } else if (fxMinterm < nvar.fxMinterm) {
    return false;
  }
  if (nextStateType) {
    int k = nextStateType.keyword();
    if (k == K_preset || k == K_clear || k == K_load) {
      return true;
    }
  }
  if (nvar.nextStateType) {
    int k = nvar.nextStateType.keyword();
    if (k == K_preset || k == K_clear || k == K_load) {
      return false;
    } 
  } 
  if (name) {
    if (name > nvar.name) { // so synchronous set/reset will come before data
      return true;
    } else {
      uassert (name < nvar.name);
      return false;
    }
  } else if (idx < nvar.idx) {
    return true;
  } else {
    uassert (idx > nvar.idx);
    return false;
  }
}

static VarInfo
getNextVar(LCell& lcell, BDD& f, int priority[], int from, int to)
{
  Cudd* mgr = f.manager();
  VarInfo best; 
  BDD bdiff, fxbar, fx;
  int* supportIdxArr = mgr->supportArray(0);
  int* suppArr1 = mgr->supportArray(1);
  int numVar = f.SupportSize(supportIdxArr);
  f.SupportIndex(supportIdxArr);

  for (int i = from; i < to; i++) {
    if (priority[i-from] > -1 && supportIdxArr[i]) {
      bdiff = f.BooleanDiff(i);
      fxbar = f.Restrict(~mgr->bddVar(i));
      fx    = f.Restrict(mgr->bddVar(i));
      VarInfo v(lcell.ncell().sequentialType() > NST_none ?
		  lcell.pinName(lcell.outPinCount()+i) : UString::Null,
		lcell.nextStateType(lcell.outPinCount()+i), i,
		bdiff.SupportSize(suppArr1), bdiff.CountMinterm(numVar-1),
		fxbar.CountMinterm(numVar-1), fx.CountMinterm(numVar-1),
		priority[i-from]);
      if (v.precedes(best)) {
	best = v;
      }
    }
  }
  return best;
}

static void
orderSupport(LCell& lcell, BDD& f, int permute[], int& orderedCount, int priority[],
	     int from, int to)
{
  Cudd* mgr     = f.manager();
  int*  suppArr = mgr->supportArray();
  int   support = f.SupportSize(suppArr);
  int   inCount = to - from;
  int   i;

  for (i = 0; i < support && orderedCount < inCount; i++) {
    VarInfo var = getNextVar(lcell, f, priority, from, to);
    if (var.idx == INT_MAX) {
      break;
    }
    uassert (orderedCount < inCount);
    permute[orderedCount++] = var.idx;
    priority[var.idx - from] = -1;
    updatePriority(f, mgr->bddVar(var.idx),priority, from, to);
  }
}

static void
setupRetentionVarsInSeqInfo(TLibSeqInfo* seqInfo, LCell& lcell)
{
  BDD f     = seqInfo->nextState();
  Cudd* mgr = f.manager();

  for (int i = 0; i < TRETPC_count; i++) {
    UPair<int,int> retPin = lcell.retentionPin((TRetPinClass)i);
    if (retPin.k_value() > 0) {
      // a valid retention pin
      int pinIdx = retPin.k_value();
      int disVal = retPin.o_value();
      int varIdx = pinIdx - lcell.outPinCount();
      BDD var = mgr->bddVar(varIdx);

      // setup the BDD for the retention anyway, regardless
      // of whether they appear in the next_state
      if (disVal == 1) {
	seqInfo->retention((TRetPinClass)i, ~var);
      } else if (disVal == 0) {
	seqInfo->retention((TRetPinClass)i, var);
      } else {
	uassert(0);
      }
    }
  }
}

static void
smoothOutRetentionVarInNextState(TLibSeqInfo* seqInfo, LCell& lcell)
{
  BDD f     = seqInfo->nextState();
  Cudd* mgr = f.manager();

  int* supportIdxArr = mgr->supportArray(0);
  int numVar = f.SupportSize(supportIdxArr);

  int retCnt = 0;
  for (int i = 0; i < TRETPC_count; i++) {
    UPair<int,int> retPin = lcell.retentionPin((TRetPinClass)i);
    if (retPin.k_value() > 0) {
      // a valid retention pin
      int pinIdx = retPin.k_value();
      int disVal = retPin.o_value();
      int varIdx = pinIdx - lcell.outPinCount();
      BDD var = mgr->bddVar(varIdx);

      if (supportIdxArr[varIdx]==1) {
	retCnt++;
	if (disVal == 1) {
	  f = f.Restrict(var);
	} else if (disVal == 0) {
	  f = f.Restrict(~var);
	} else {
	  uassert(0);
	}
      }
    }
  }

  if (retCnt > 0) {
    if (f.SupportSize(supportIdxArr) != (numVar-retCnt)) {
      uassert(0);
    }
    seqInfo->nextState(f);
  }
}

static void
orderInPinsBasedOnSeqInfo(LCell& cell, int permute[], int& orderedCount,
			  int priority[], int inCount)
{
  int  from     = cell.bidirPinCount();
  int  to       = from+inCount;

  // setup the retention BDDs in seqInfo before reorder
  if (cell.hasMVInfo()) {
    for (UAggIter iter = cell.internalPins().msbToLsb();
	 iter; iter++) {
      TLibPin tpin = cell.pin(iter);
      if (tpin.isInverted()) {
	continue;
      }
      TLibSeqInfo* seqInfo = tpin.seqInfo();
      // this setup retention BDD on seqInfo->retention(i)
      setupRetentionVarsInSeqInfo(seqInfo, cell);
      // this remove retention vars from next_state, if any
      if (seqInfo->nextState().ptr()) {
	smoothOutRetentionVarInNextState(seqInfo, cell);
      }
    }
  }

  for (UAggIter iter = cell.internalPins().msbToLsb();
       iter && orderedCount < inCount; iter++) {
    TLibPin tpin = cell.pin(iter);
    if (tpin.isInverted()) {
      continue;
    }
    TLibSeqInfo* seqInfo = tpin.seqInfo();
    BDD f;

    f = seqInfo->nextState();

    if (f.ptr()) {
      orderSupport(cell, f, permute, orderedCount, priority, from, to);
      if (orderedCount == inCount) break;
    } // else probably an RS_latch

    f = seqInfo->clock();
    if (f.ptr()) {
      orderSupport(cell, f, permute, orderedCount, priority, from, to);
      if (orderedCount == inCount) break;
    } // else probably an RS_latch

    f = seqInfo->clear();
    if (f.ptr()) {
      orderSupport(cell, f, permute, orderedCount, priority, from, to);
      if (orderedCount == inCount) break;
    }

    f = seqInfo->preset();
    if (f.ptr()) {
      orderSupport(cell, f, permute, orderedCount, priority, from, to);
      if (orderedCount == inCount) break;
    }

    if (seqInfo->type() > NST_latch) {
      TLibMSFFInfo* ms = static_cast<TLibMSFFInfo*>(seqInfo);
      f = ms->slaveClock();
      orderSupport(cell, f, permute, orderedCount, priority, from, to);
    }

    if (cell.hasMVInfo()) {
      for (int i = 0; i < TRETPC_count; i++) {
	f = seqInfo->retention((TRetPinClass)i);
	if (f.ptr()) { 
	  uassert(cell.retentionPin((TRetPinClass)i).k_value() > 0);
	  orderSupport(cell, f, permute, orderedCount, priority, from, to);
	  if (orderedCount == inCount) break;
	}
      }
    }
  }
}

static void
orderInPinsBasedOnDriverPins(LCell& cell, int permute[], int& orderedCount,
			     int priority[], int inCount, Cudd& mgr)
{
  int   from         = cell.bidirPinCount();
  int   to           = from+inCount;
  int   newIdxOffset = cell.internalPinCount()+to;
  BDD   f            = mgr.bddZero(); 
  int   newIdx       = newIdxOffset;

  for (UAggIter iter = cell.driverPins().msbToLsb();
       iter && orderedCount < inCount; iter++) {
    TLibPin tpin = cell.pin(iter);
    f ^= tpin.function() * mgr.bddVar(newIdx++); 
  }
  orderSupport(cell, f, permute, orderedCount, priority, from, to);
  if (orderedCount == inCount) {
    return;
  }
  // now canonicalize based on the tristate functions if any exist
  f = mgr.bddZero();
  for (UAggIter iter = cell.driverPins().msbToLsb();
       iter && orderedCount < inCount; iter++) {
    TLibPin tpin = cell.pin(iter);
    if (tpin.isTristateDriver()) {
      f ^= tpin.tristate() * mgr.bddVar(newIdx++); 
    }
  }
  if (f != mgr.bddZero()) {
    orderSupport(cell, f, permute, orderedCount, priority, from, to);
  }
}

static void
permuteFunctions(LCell& cell, int permute[], Cudd& mgr)
{
  int   inCount    = cell.inPinCount();
  int   bidirCount = cell.bidirPinCount();
  int   size       = mgr.ReadSize();
  int*  newPermute = new int[size];
  int   i;

  for (i = 0; i < bidirCount; i++) {
    newPermute[i] = i;
  }
  for (i = 0; i < inCount; i++) {
    newPermute[permute[i]] = i+bidirCount;
  }
  for (i = bidirCount+inCount; i < size; i++) {
    newPermute[i] = i;
  }

  for (UAggIter iter = cell.driverPins().msbToLsb(); iter; iter++) {
    TLibPin tpin = cell.pin(iter);
    tpin.function(tpin.function().Permute(newPermute));
    if (tpin.isTristateDriver()) {
      tpin.tristate(tpin.tristate().Permute(newPermute));
    }
  }
      
  for (UAggIter iter = cell.internalPins().msbToLsb(); iter; iter++) {
    TLibPin tpin = cell.pin(iter);
    if (tpin.isInverted() || tpin.isTestInternal() /*test_internal is processed later*/) {
      continue;
    }
    TLibSeqInfo* seq = tpin.seqInfo();
    if (seq->nextState().ptr()) {
      seq->nextState(seq->nextState().Permute(newPermute));
    }
    if (seq->clock().ptr()) {
      seq->clock(seq->clock().Permute(newPermute));
    }
    if (seq->preset().ptr()) {
      seq->preset(seq->preset().Permute(newPermute));
    }
    if (seq->clear().ptr()) {
      seq->clear(seq->clear().Permute(newPermute));
    }
    if (seq->type() > NST_latch) {
      TLibMSFFInfo* ms = static_cast<TLibMSFFInfo*>(seq);
      ms->slaveClock(ms->slaveClock().Permute(newPermute));
    }
    for (int r = 0; r < TRETPC_count; r++) {
      TRetPinClass pc = (TRetPinClass)r;
      if (seq->retention(pc).ptr()) {
	seq->retention(pc,seq->retention(pc).Permute(newPermute));
      }
    }
  }
  delete [] newPermute;
}

static void
permuteTestCellFunctions(LCell& cell, int permute[], Cudd& mgr)
{
  int   inCount    = cell.inPinCount();
  int   bidirCount = cell.bidirPinCount();
  int   size       = mgr.ReadSize();
  int*  newPermute = new int[size];
  int   i;

  for (i = 0; i < bidirCount; i++) {
    newPermute[i] = i;
  }
  // first order the non-test input pins
  int   offset     = cell.outPinCount();
  int   cnt        = bidirCount;
  for (i = 0; i < inCount; i++) {
    int idx = i + offset;
    if (cell.pin(idx).signalType() == TLPST_none) {
      // not a test pin
      newPermute[permute[i]] = cnt++;
    }
  }
  // create pinMap to annotate to the test internal pin
  int pinMapSize = cnt - bidirCount;
  int* pinMap = new IN_SAME_POOL_AS(cell.pin(bidirCount).ptr()) int[pinMapSize+1];
  int j = 0;
  pinMap[j++] = pinMapSize;
  for (i = 0; i < inCount; i++) {
    int idx = i + offset;
    if (cell.pin(idx).signalType() == TLPST_none) {
      // not a test pin
      pinMap[j++] = i;
    }
  }
  // order the internal variables of the scan core next
  i = bidirCount+inCount;
  for (UAggIter iter = cell.internalPins().msbToLsb(); iter && i < size; iter++) {
    TLibPin tpin = cell.pin(iter);
    if (tpin.isTestInternal()) {
      newPermute[i++] = cnt++;
    }
  }
  // now order the rest - actually the order of these should be inconsequential
  for (UAggIter iter = cell.internalPins().msbToLsb(); iter && i < size; iter++) {
    TLibPin tpin = cell.pin(iter);
    if (!tpin.isTestInternal()) {
      newPermute[i++] = cnt++;
    }
  }
  for (i = 0; i < inCount; i++) {
    int idx = i + offset;
    if (cell.pin(idx).signalType() != TLPST_none) {
      // test pin
      newPermute[permute[i]] = cnt++;
    }
  }
  for (i = bidirCount+inCount+cell.internalPins().count(); i < size; i++) {
    newPermute[i] = cnt++;
  }

  for (UAggIter iter = cell.driverPins().msbToLsb(); iter; iter++) {
    TLibPin tpin = cell.pin(iter);
    if (tpin.functionForScanCore().ptr()) {
      tpin.functionForScanCore(tpin.functionForScanCore().Permute(newPermute));
    }
  }
      
  for (UAggIter iter = cell.internalPins().msbToLsb(); iter; iter++) {
    TLibPin tpin = cell.pin(iter);
    if (tpin.isInverted() || !tpin.isTestInternal()) {
      continue;
    }
    // there should never be more than one test_internal pin
    uassert(pinMap);
    tpin.pinMapForTest(pinMap);
    pinMap = NULL;
    TLibSeqInfo* seq = tpin.seqInfo();
    if (seq->nextState().ptr()) {
      seq->nextState(seq->nextState().Permute(newPermute));
    }
    if (seq->clock().ptr()) {
      seq->clock(seq->clock().Permute(newPermute));
    }
    if (seq->preset().ptr()) {
      seq->preset(seq->preset().Permute(newPermute));
    }
    if (seq->clear().ptr()) {
      seq->clear(seq->clear().Permute(newPermute));
    }
    if (seq->type() > NST_latch) {
      TLibMSFFInfo* ms = static_cast<TLibMSFFInfo*>(seq);
      ms->slaveClock(ms->slaveClock().Permute(newPermute));
    }
    for (int r = 0; r < TRETPC_count; r++) {
      TRetPinClass pc = (TRetPinClass)r;
      if (seq->retention(pc).ptr()) {
	seq->retention(pc,seq->retention(pc).Permute(newPermute));
      }
    }
  }
  delete [] newPermute;
}

static void
orderScanInputs(LCell& cell, int permute[], int orderedCount,
		int priority[], int inCount)
{
  UList<int> priorityList;
  UList<int>::iterator it;
  UAggIter piter = cell.inPins().msbToLsb();
  int offset = piter.current();
  for (; piter; ++piter) {
    int idx = piter.current() - offset;
    if (priority[idx] < 0) {
      continue;
    }
    switch (cell.pin(piter).signalType()) {
    case TLPST_test_scan_enable:
      priority[idx] = 1;
      break;
    case TLPST_test_scan_enable_inverted:
      priority[idx] = 2;
      break;
    case TLPST_test_scan_clock:
      priority[idx] = 3;
      break;
    case TLPST_test_scan_clock_a:
      priority[idx] = 4;
      break;
    case TLPST_test_scan_clock_b:
      priority[idx] = 5;
      break;
    case TLPST_test_clock:
      priority[idx] = 6;
      break;
    case TLPST_test_scan_in:
      priority[idx] = 7;
      break;
    case TLPST_test_scan_in_inverted:
      priority[idx] = 8;
      break;
    default:
      continue;
    }
    // insert idx into priority list
    int p = priority[idx];
    for (it = priorityList.begin(); it; ++it) {
      if (p <= priority[it.object()]) {
	break;
      }
    }
    priorityList.insertBefore(it, idx);
  }
  uassert ((orderedCount + (int)priorityList.size()) <= inCount);
  offset = cell.bidirPinCount();
  for (it = priorityList.begin(); it; ++it) {
    int idx = it.object();
    priority[idx] = -1;
    permute[orderedCount++] = idx+offset;
  }
}

static void
canonicalizeInPinOrder(LCell& cell, NSeqType seqType, Cudd& mgr)
{
  // should probably move this to canonicalizePinOrder,
  // but needs to check first if any function is relying on the
  // NULL drivePins being first
  if (cell.inPinCount() < 2) {
    return;
  }
  int* permute;
  int* priority;
  int  orderedCount = 0,   // number of inputs processed
       idxOffset    = cell.bidirPinCount(),
       inCount      = cell.inPinCount();
  int  i;

  // check last internal pin to see if cell is a scan cell
  UAggIter intPinIter = cell.internalPins().lsbToMsb();
  bool isScanCell = (intPinIter && cell.pin(intPinIter).isTestInternal());

  permute = new int[inCount];
  priority = new int[inCount];
  for (i=0; i<inCount; i++) {
    priority[i] = 0;
  }
  
  if (seqType > NST_none) {
    orderInPinsBasedOnSeqInfo(cell, permute, orderedCount, priority,
			      inCount); 
    // the following takes care of the situation when there is
    // no ff or latch definition at the scan cell level, but
    // only at the core or non-scan level, in which case, the
    // scan inputs will not have been ordered
    if (isScanCell && orderedCount < inCount) {
      orderScanInputs(cell, permute, orderedCount, priority, inCount);
    }
  }
  
  if (orderedCount < inCount) {
    orderInPinsBasedOnDriverPins(cell, permute, orderedCount, priority,
				 inCount, mgr);
  }

  //  order remaining unused inputs
  for (i = 0; i < inCount && priority[i] == -1; i++);
  int start = i;
  while (orderedCount < inCount) {
    // the following is not efficient in general, but most of the time,
    // we will either be dealing with a small set of variables here
    // or should not have to do this at all
    int firstPriority = -1;
    int first = -1;
    for (i = start; i < inCount; i++) {
      if (firstPriority < priority[i]) {
	firstPriority = priority[i];
	first = i;
      }
    }
    if (first > -1) {
      permute[orderedCount++] = first+idxOffset;
      priority[first] = -1;
    } else {
      break;
    }
  }
  for (i = 0; i < orderedCount && permute[i] == i+idxOffset; i++);
  if (i != orderedCount) {
    // need to reorder pins and permute variables in BDDs
    cell.permutePins(cell.inPins(), permute, idxOffset);
    permuteFunctions(cell, permute, mgr);
  }
  if (isScanCell) {
    permuteTestCellFunctions(cell, permute, mgr);
  }
  delete [] priority;
  delete [] permute;
}

static void
canonicalizePinOrderBasedOnSignalType(LCell& cell, const UAggSlice& pins,
				      TLibPinSignalType signalTypes[])
{
  int start = pins.offset();
  int end   = start + pins.count();
  for (int i = start; i < end; ++i) {
    bool found = false;
    for (int j = i; j < end; ++j) {
      TLibPin& pin = cell.pin(j);
      if (pin.signalType() != signalTypes[i-start]) {
	continue;
      } else {
	found = true;
	if (j != i) {
	  cell.swapPins(i, j);
	}
	break;
      }
    }
  }
}

static void
canonicalizePinOrderForClockGate(LCell& cell)
{
  TLibPinSignalType inType[3] = {TLPST_clock_gate_enable, TLPST_clock_gate_clock,
				 TLPST_clock_gate_test};
  TLibPinSignalType outType[2] = {TLPST_clock_gate_out, TLPST_clock_gate_obs};

  canonicalizePinOrderBasedOnSignalType(cell, cell.inPins(), inType);
  canonicalizePinOrderBasedOnSignalType(cell, cell.outPins(), outType);
}  

inline static void
canonicalizePinOrder(LCell& cell, NSeqType seqType, Cudd& mgr)
{
  if (cell.ncell().isClockGate()) {
    canonicalizePinOrderForClockGate(cell);
  } else if (!pointless2Canonicalize(cell)) {
    canonicalizeInPinOrder(cell, seqType, mgr);
    canonicalizeDrivePinOrder(cell, ((seqType > NST_none) ? getNodeIndex : getFunctionPtr));
  }
}
  
static inline void
addInternalPins(const LibGroup* group, NCell cell, int cnt = 1)
{
  cell.addInternalPin(group->numberOfNames() * cnt);
}

static inline int
numberOfPins(const LibGroup* group)
{
  return group->numberOfNames();
}

static inline int
numberOfPGPins(const LibGroup* group)
{
  return 1;
}

static LibAttr*
getGroupDirection(const LibGroup* g)
{
  LibErrorT err;
  LibAttr* ldir = g->findAttr(K_direction, err);
  if (!ldir && g->typeID() != LIBERTY_GROUPENUM_pg_pin) {
    // direction not defined - must be a bus or bundle
    // with pin groups defined
    uassert (g->typeID() == LIBERTY_GROUPENUM_bus ||
	     g->typeID() == LIBERTY_GROUPENUM_bundle);
    // find first pin definition with direction
    LibGroupIter giter = g->groupIter();
    LibGroup* p;
    while ((p = giter++)) {
      if (p->typeID() != LIBERTY_GROUPENUM_pin) {
	continue;
      }
      ldir = p->findAttr(K_direction, err);
      if (ldir) {
	break;
      }
    }
  }
  return ldir;
}

static LibAttr*
getGroupPGType(const LibGroup* g)
{
  LibErrorT err;
  LibAttr* ldir = g->findAttr(K_pg_type, err);
  return ldir;
}

static LibAttr*
isGroupThreeState(const LibGroup* g)
{
  LibErrorT err;
  LibAttr* attr = g->findAttr("three_state", err);
  if (!attr) {
    uassert (g->typeID() == LIBERTY_GROUPENUM_bus ||
	     g->typeID() == LIBERTY_GROUPENUM_bundle);
    // find first pin definition with direction
    LibGroupIter giter = g->groupIter();
    LibGroup* p;
    while ((p = giter++)) {
      if (p->typeID() != LIBERTY_GROUPENUM_pin) {
	continue;
      }
      attr = p->findAttr("three_state", err);
      if (attr) {
	break;
      }
    }
  }
  return attr;
}

static void
updateNumPins(const LibGroup* lpin, int numPins, int& output, int& bidir, int& input,
	      int& internal)
{
  LibAttr* ldir = getGroupDirection(lpin);
  TLibDir  tdir;
  if (!ldir) {
    if (lpin->typeID() == LIBERTY_GROUPENUM_pg_pin) {
      tdir = TLD_input; // default PG direction
    } else {
      uassert(0);
      tdir = TLD_input;
    }
  } else {
    tdir = (TLibDir)UString(ldir->value()->string_val()).keyword();
  }
  switch (tdir) {
  case TLD_tristate:
  case TLD_output: output += numPins; break;
  case TLD_bidir: bidir += numPins; break;
  case TLD_input: input += numPins; break;
  case TLD_internal: internal += numPins; break;
  }
}

void
LLoad::countPins(const LibGroup* lcell, LCell& tmpCell, 
		 int& output, int& bidir, int& input, int& internal,
		 int& pgOutput, int& pgBidir, int& pgInput, int& pgInternal,
		 LibGroup*& testCell, LibGroup*& coreCell)
{
  LibErrorT err;
  LibGroupIter giter;
  LibGroup* lpin;
  int numPins;

  testCell = NULL;
  coreCell = NULL;
  output = bidir = input = internal = 0;
  pgOutput = pgBidir = pgInput = pgInternal = 0;

  giter = lcell->groupIter();
  while ((lpin = giter++)) {
    switch (lpin->typeID()) {
    case LIBERTY_GROUPENUM_ff:
    case LIBERTY_GROUPENUM_latch:
      internal += numberOfPins(lpin);
      break;
    case LIBERTY_GROUPENUM_test_cell:
      if (testCell) {
	LIBLOADMSG(multipleTestCells, (const char*)tmpCell.ncell().name());
      } else {
	testCell = lpin;
      }
      break;
    case LIBERTY_GROUPENUM_bundle :
      {
	LibAttr* mem = lpin->findAttr("members", err);
	numPins = mem->valueCount();
      }
      updateNumPins(lpin, numPins, output, bidir, input, internal);
      break;
    case LIBERTY_GROUPENUM_pg_pin :
      numPins = numberOfPGPins(lpin);
      updateNumPins(lpin, numPins, pgOutput, pgBidir, pgInput, pgInternal);
      break;
    case LIBERTY_GROUPENUM_pin :
      numPins = numberOfPins(lpin);
      updateNumPins(lpin, numPins, output, bidir, input, internal);
      break;
    case LIBERTY_GROUPENUM_bus:
      numPins = numberOfPins(lpin);
      {
	LibAttr* busName = lpin->findAttr("bus_type", err);
	UAggType* busType = findType(busName->value()->string_val(), tmpCell);
	uassert (busType);
	numPins = numPins * busType->size();
      }
      updateNumPins(lpin, numPins, output, bidir, input, internal);
      break;
    default :
      break;
    }
  }
  if (testCell) {
    giter = testCell->groupIter();
    while ((lpin = giter++)) {
      switch (lpin->typeID()) {
      case LIBERTY_GROUPENUM_ff:
      case LIBERTY_GROUPENUM_latch:
	coreCell = lpin;
	internal += numberOfPins(lpin);
	return;
      default:
	continue;
      }
    }
  }
}

void
LLoad::getPinMap(const LibGroup* lcell, const LibGroup* scanCore,
		 LCell& tmpCell)
{
  LLoadPinMap& pinMap = tmpCell.pinMap();
  LibErrorT err;
  LibGroupIter giter;
  LibGroup* lpin;
  LibNameIter niter;
  UString name;
  LibAttr* ldir;
  TLibDir tdir;
  UAggIter inPinIter = tmpCell.inPins().msbToLsb();
  UAggIter bidirPinIter = tmpCell.bidirPins().msbToLsb();
  UAggIter outPinIter = tmpCell.outPins().msbToLsb();
  UAggIter intPinIter = tmpCell.internalPins().msbToLsb();
  UAggIter* iter = NULL;

  giter = lcell->groupIter();
  while ((lpin = giter++)) {
    switch (lpin->typeID()) {
    case LIBERTY_GROUPENUM_ff:
    case LIBERTY_GROUPENUM_latch:
      {
	niter = lpin->nameIter();
	while ((name = niter++)) {
	  tmpCell.setPinName(intPinIter.current(), name);
	  pinMap[name] = intPinIter++.current();
	}
      }
      break;
    case LIBERTY_GROUPENUM_pin:
    case LIBERTY_GROUPENUM_bus:
    case LIBERTY_GROUPENUM_bundle:
      {
	LibAttr* busName;
	UAggType* busType = NULL;
	if (lpin->typeID() == LIBERTY_GROUPENUM_bus) {
	  busName = lpin->findAttr("bus_type", err);
	  busType = findType(busName->value()->string_val(), tmpCell);
	  uassert (busType);
	}

	ldir = getGroupDirection(lpin);
	tdir = (TLibDir)UString(ldir->value()->string_val()).keyword();
	switch (tdir) {
	case TLD_tristate:
	case TLD_output: iter = &outPinIter; break;
	case TLD_bidir: iter = &bidirPinIter; break;
	case TLD_input: iter = &inPinIter; break;
	case TLD_internal: iter = &intPinIter; break;
	}
	niter = lpin->nameIter();
	while ((name = niter++)) {
	  // for busses the following will register a map of the bus name
	  // to the first pin in the bus
	  if (lpin->typeID() == LIBERTY_GROUPENUM_pin) {
	    tmpCell.setPinName((*iter).current(), name);
	  }
	  pinMap[name] = (*iter).current();
	  if (lpin->typeID() == LIBERTY_GROUPENUM_bundle) {
	    LibAttr* mem = lpin->findAttr("members", err);
	    tmpCell.bundles()[name] = mem->valueCount();
	    for (LibValIter viter = mem->valueIter(); viter; ++viter) {
	      tmpCell.setPinName((*iter).current(), viter.object()->string_val());
	      pinMap[viter.object()->string_val()] = (*iter)++.current();
	    }
	  } else if (busType) {
	    tmpCell.busses()[name] = busType;
	    for (int i = 0; i < busType->size(); i++) {
	      UStrBuf index = busType->name(i, UNM_vlog1995);
	      UStrBuf component = name;
	      component << index;
	      tmpCell.setPinName((*iter).current(), UString(component));
	      pinMap[UString(component)] = (*iter)++.current();
	    }
	  } else {
	    (*iter)++;
	  }
	}
      }
      break;
    default:
      break;
    }
  }
  if (scanCore) {
    // map the internal pins of the scan core to the same
    // internal index of the wrapper for function matching
    intPinIter = tmpCell.internalPins().msbToLsb();
    uassert (scanCore->typeID() == LIBERTY_GROUPENUM_ff ||
	     scanCore->typeID() == LIBERTY_GROUPENUM_latch);
    niter = scanCore->nameIter();
    while ((name = niter++)) {
      // add the internal names only if they are different
      // from those used in the scan wrapper
      if (pinMap.find(name)) {
	// if the same internal name is used in the core
        // as in the wrapper, they must be in the same order

	//	uassert (pinMap[name] == intPinIter.current());
	++intPinIter;
      } else {
	tmpCell.setPinName(intPinIter.current(), name);
	pinMap[name] = intPinIter++.current();
      }
    }
  }
}

void
LLoad::processInternalSeqPins(const LibGroup* seq, bool isTest, UAggIter& intPinIter,
			      LCell& lcell)
{
  LLoadPinMap& pinMap = lcell.pinMap();
  int outCount = lcell.outPinCount();
  TLibPin tpin(d_lib.tlib(), TLD_internal, lcell.totalPinCount(), isTest);
  LibNameIter niter = seq->nameIter();
  UString name;
  bool first = true;
  while ((name = niter++)) {
    int idx = intPinIter++.current();
    lcell.pin(idx).grab(tpin);
    if (!first) {
      lcell.pin(idx).invert(); // second pin is inverted
    } else {
      first = false;
    }
  }
  processSeqInfo(seq, (TLibInternalPinC*)tpin.ptr(), seq->typeID(),
		 lcell.ncell(), outCount, pinMap);
}

void
LLoad::processBundlePinAttributes(const LibGroup* bundle, LCell& lcell,
				  UAggIter& outPinIter, UAggIter& bidirPinIter,
				  UAggIter& inPinIter, UAggIter& intPinIter)
{
  LLoadPinMap& pinMap = lcell.pinMap();
  LLoadBundleMap& bundles = lcell.bundles();
  int outCount = lcell.outPinCount();
  LibErrorT err;
  LibAttr* ldir = getGroupDirection(bundle);
  TLibDir  tdir = (TLibDir)UString(ldir->value()->string_val()).keyword();
  UAggIter* iter = NULL;
  
  LibAttr* mem = bundle->findAttr("members", err);
  LibValIter viter;
  int bundleIdx;
  for (viter = mem->valueIter(), bundleIdx = 0; viter; ++viter, ++bundleIdx) {
    if (tdir == TLD_output && isGroupThreeState(bundle)) {
      // is a tristate function
      tdir = TLD_tristate;
    }
    TLibPin tpin(d_lib.tlib(), tdir, (uint_t)lcell.totalPinCount(), /*scan=*/false);
    switch (tdir) {
    case TLD_tristate:
    case TLD_output: iter = &outPinIter; break;
    case TLD_bidir: iter = &bidirPinIter; break;
    case TLD_input: iter = &inPinIter; break;
    case TLD_internal: 
      {
	UString nm = viter.object()->string_val();
	tpin.setInternalPinName(nm);
	iter = &intPinIter;
      }
      break;
    }
    int idx = (*iter)++.current();
    lcell.pin(idx).grab(tpin);
    // process bundle attributes for pin
    if (tdir == TLD_internal) {
      // no attribute processing for internal pins
      continue;
    }
    processPinAttr(bundle, (TLibExternalPinC*)tpin.ptr(), tdir,
		   lcell, idx,
		   outCount, pinMap, &bundles, NULL, NULL, bundleIdx);
    const LibGroup* lpin = bundle->findGroup(viter.object()->string_val(), "pin", err);
    if (lpin) {
      // if individual pin definition exists, overwrite bundle definition
      processPinAttr(lpin, (TLibExternalPinC*)tpin.ptr(), tdir,
		     lcell, idx,
		     outCount, pinMap, NULL, NULL, NULL, -1);
    }
  }
}

void
LLoad::processBundlePinSubgroup(const LibGroup* bundle, LCell& lcell, uint_t& delayCount)
{
  LLoadPinMap& pinMap = lcell.pinMap();
  LLoadBundleMap& bundles = lcell.bundles();
  LibErrorT err;

  LibAttr* mem = bundle->findAttr("members", err);
  LibValIter viter;
  int bundleIdx;
  for (viter = mem->valueIter(), bundleIdx = 0; viter; ++viter, ++bundleIdx) {
    UString name = viter.object()->string_val();
    int tidx = pinMap[name];
    // for now, we only process timing groups within the pin group;
    // we take it either from the individual pin definition or the bundle definition
    const LibGroup* lpin = bundle->findGroup(name, "pin", err);
    if (lpin) {
      processPinSubgroup(lpin, tidx, lcell, pinMap, NULL, NULL, NULL, -1, delayCount, /*dontAddIfExist=*/false);
    }
    processPinSubgroup(bundle, tidx, lcell, pinMap, &bundles, NULL, NULL, bundleIdx, delayCount, /*dontAddIfExist=*/true);
  }
}

// process individual pin statements within the bus group
// this will overwrite the default from the bus group
void
LLoad::processBusPinAttributes(const LibGroup* bus, LCell& lcell,
			       UAggIter& outPinIter, UAggIter& bidirPinIter,
			       UAggIter& inPinIter, UAggIter& intPinIter,
			       const LibGroup* testCell)
{
  LLoadPinMap& pinMap = lcell.pinMap();
  LLoadBusMap& busses = lcell.busses();
  int outCount = lcell.outPinCount();
  LibErrorT err;
  UString name;
  UAggIter* iter = NULL;
  TLib* tlib = d_lib.tlib();
  
  // process the bus attributes for all the pins in the bus definition
  LibAttr*  busName = bus->findAttr("bus_type", err);
  UAggType* busType = findType(busName->value()->string_val(), lcell);
  uassert (busType);
  LibAttr*  ldir = getGroupDirection(bus);
  TLibDir   tdir = (TLibDir)UString(ldir->value()->string_val()).keyword();
  if (tdir == TLD_output && isGroupThreeState(bus)) {
    // is a tristate function
    tdir = TLD_tristate;
  }
  switch (tdir) {
  case TLD_tristate:
  case TLD_output: iter = &outPinIter; break;
  case TLD_bidir: iter = &bidirPinIter; break;
  case TLD_input: iter = &inPinIter; break;
  case TLD_internal: iter = &intPinIter; break;
  }
  LibNameIter niter = bus->nameIter();
  while ((name = niter++)) {
    int num = busType->size();
    for (int i = 0; i < num; i++) {
      TLibPin tpin(tlib, tdir, (uint_t)lcell.totalPinCount(), testCell);
      int idx = (*iter)++.current();
      lcell.pin(idx).grab(tpin);
      // process pin function and other attributes
      if (tdir == TLD_internal) {
	UStrBuf nm;
	nm << name << '[' << i << ']';
	tpin.setInternalPinName((const char*)nm);
	// no attribute processing for internal pins
	continue;
      }
      processPinAttr(bus, (TLibExternalPinC*)tpin.ptr(), tdir,
		     lcell, idx, outCount, pinMap, /*bundleMap=*/NULL,
		     &busses, busType, i);
    }
  }
  
  if (tdir == TLD_internal) {
    // no attribute processing for internal pins
    return;
  }
  // now process the individual pin definitions within the bus group
  // to overwrite the bus attributes
  LibGroup* lpin;
  LibGroupIter giter = bus->groupIter();
  while ((lpin = giter++)) {
    switch (lpin->typeID()) {
    case LIBERTY_GROUPENUM_pin:
      {
	niter = lpin->nameIter();
	while ((name = niter++)) {
	  name = UNaming::descapeForLiberty(name);
	  if (pinMap.find(name)) {
	    int idx = pinMap[name];
	    TLibPin tpin = lcell.pin(idx);
	    // process pin function and other attributes
	    processPinAttr(lpin, (TLibExternalPinC*)tpin.ptr(), tdir,
			   lcell, idx,
			   outCount, pinMap, /*bundleMap=*/NULL,
			   /*busMap=*/NULL, /*busType*/NULL, /*grpIdx=*/-1);
	  } else {
	    // must be a name with a range
	    int from, to;
	    char nm[1000];
	    int idx = sscanf(name.str(), "%[^[][%d:%d]", nm, &from, &to);
	    uassert (idx == 3);
	    if (from > to) {
	      int tmp = to;
	      to = from;
	      from = tmp;
	    }
	    for (int i = from; i <= to; ++i) {
	      UStrBuf str(nm);
	      str << "[" << i << "]";
	      uassert (pinMap.find((const char*)str));
	      idx = pinMap[(const char*)str];
	      TLibPin tpin = lcell.pin(idx);
	      // process pin function and other attributes
	      processPinAttr(lpin, (TLibExternalPinC*)tpin.ptr(), tdir,
			     lcell, idx,
			     outCount, pinMap, /*bundleMap=*/NULL,
			     /*busMap=*/NULL, /*busType*/NULL, /*grpIdx=*/-1);
	    }
	  }
	}
      }
      break;
    default:
      break;
    }
  }
}

void
LLoad::processBusPinSubgroup(const LibGroup* bus, LCell& lcell, uint_t& delayCount)
{
  LLoadPinMap& pinMap = lcell.pinMap();
  LLoadBusMap& busses = lcell.busses();
  LibErrorT err;
  LibGroup* lpin;
  UString name;
  
  // for subgroups of bus/pins, we only process the timing group today;
  // we first process individual pin definition first if they exist
  LibGroupIter giter = bus->groupIter();
  while ((lpin = giter++)) {
    switch (lpin->typeID()) {
    case LIBERTY_GROUPENUM_pin:
      {
	LibNameIter niter = lpin->nameIter();
	while ((name = niter++)) {
	  name = UNaming::descapeForLiberty(name);
	  if (pinMap.find(name)) {
	    int idx = pinMap[name];
	    // process timing groups and other subgroups of pin
	    // this must be done after pins have been reordered
	    processPinSubgroup(lpin, idx, lcell, pinMap, NULL, NULL, NULL, -1, delayCount,
			       /*dontAddIfExist=*/false);
	  } else {
	    // must be a name with a range
	    int from, to;
	    char nm[1000];
	    int idx = sscanf(name.str(), "%[^[][%d:%d]", nm, &from, &to);
	    uassert (idx == 3);
	    if (from > to) {
	      int tmp = to;
	      to = from;
	      from = tmp;
	    }
	    for (int i = from; i <= to; ++i) {
	      UStrBuf str(nm);
	      str << "[" << i << "]";
	      uassert (pinMap.find((const char*)str));
	      idx = pinMap[(const char*)str];
	      processPinSubgroup(lpin, idx, lcell, pinMap, NULL, NULL, NULL, -1, delayCount,
				 /*dontAddIfExist=*/false);
	    }
	  }
	}
      }
      break;
    default:
      break;
    }
  }

  // for any pin in the bus that has not been individually defined,
  // we take the bus definition for that pin
  LibAttr*  busName = bus->findAttr("bus_type", err);
  UAggType* busType = findType(busName->value()->string_val(), lcell);
  uassert (busType);
  int num = busType->size();

  LibNameIter niter = bus->nameIter();
  while ((name = niter++)) {
    uassert (pinMap.find(name));
    int idx = pinMap[name];
    for (int i = 0; i < num; i++) {
      int tidx = idx+i;
      TLibPin tpin(lcell.pin(tidx));
      processPinSubgroup(bus, tidx, lcell, pinMap, /*bundles=*/NULL,
			 &busses, busType, i, delayCount, /*dontAddIfExist=*/true);
    }
  }
}

void
LLoad::processPGPins(const LibGroup* cellGroup,
		     LCell& lcell)
{
  LibErrorT err;
  LibGroupIter giter;
  LibGroup* lpin;
  LibNameIter niter;
  UString name;
  LibAttr* ldir;
  TLibDir tdir;
  UAggIter pginPinIter = lcell.inPGPins().msbToLsb();
  UAggIter pgbidirPinIter = lcell.bidirPGPins().msbToLsb();
  UAggIter pgoutPinIter = lcell.outPGPins().msbToLsb();
  UAggIter pgintPinIter = lcell.internalPGPins().msbToLsb();
  UAggIter* pgiter = NULL;

  giter = cellGroup->groupIter();
  while ((lpin = giter++)) {
    switch (lpin->typeID()) {
    case LIBERTY_GROUPENUM_pg_pin:
      {
	ldir = lpin->findAttr(K_direction, err);
	if (ldir) {
	  tdir = (TLibDir)UString(ldir->value()->string_val()).keyword();
	} else {
	  tdir = TLD_input;
	}

	switch (tdir) {
	case TLD_tristate:
	case TLD_output: pgiter = &pgoutPinIter; break;
	case TLD_bidir: pgiter = &pgbidirPinIter; break;
	case TLD_input: pgiter = &pginPinIter; break;
	case TLD_internal: pgiter = &pgintPinIter; break;
	}
	niter = lpin->nameIter();
	while ((name = niter++)) {
	  TLibPGPin tpin(d_lib.tlib(), name, tdir);
	  int idx = (*pgiter)++.current();
	  lcell.pgpin(idx)->grab(tpin);
	  processPGPinAttr(lpin, tpin.ptr(), tdir,
			   lcell, idx);
	}
      }
      break;
    default:
      break;
    }
  }
}

void
LLoad::processPins(const LibGroup* cellGroup, const LibGroup* testCell,
		   LCell& lcell)
{
  LLoadPinMap& pinMap = lcell.pinMap();
  LLoadBusMap& busses = lcell.busses();
  LibErrorT err;
  LibGroupIter giter;
  LibGroup* lpin;
  LibNameIter niter;
  UString name;
  LibAttr* ldir;
  TLibDir tdir;
  UAggIter inPinIter = lcell.inPins().msbToLsb();
  UAggIter bidirPinIter = lcell.bidirPins().msbToLsb();
  UAggIter outPinIter = lcell.outPins().msbToLsb();
  UAggIter intPinIter = lcell.internalPins().msbToLsb();
  UAggIter* iter = NULL;
  int outCount = lcell.outPinCount();

  giter = cellGroup->groupIter();
  while ((lpin = giter++)) {
    switch (lpin->typeID()) {
    case LIBERTY_GROUPENUM_ff:
    case LIBERTY_GROUPENUM_latch:
      processInternalSeqPins (lpin, /*isTest=*/false, intPinIter,
			      lcell);
      break;
    case LIBERTY_GROUPENUM_bundle:
      processBundlePinAttributes(lpin, lcell, outPinIter, bidirPinIter, inPinIter, intPinIter);
      break;
    case LIBERTY_GROUPENUM_bus:
      processBusPinAttributes(lpin, lcell, outPinIter, bidirPinIter, inPinIter, intPinIter, testCell);
      break;
    case LIBERTY_GROUPENUM_pin:
      {
	ldir = lpin->findAttr(K_direction, err);
	tdir = (TLibDir)UString(ldir->value()->string_val()).keyword();
	if (tdir == TLD_output && lpin->findAttr("three_state", err)) {
	  // is a tristate function
	  tdir = TLD_tristate;
	}
	switch (tdir) {
	case TLD_tristate:
	case TLD_output: iter = &outPinIter; break;
	case TLD_bidir: iter = &bidirPinIter; break;
	case TLD_input: iter = &inPinIter; break;
	case TLD_internal: iter = &intPinIter; break;
	}
	niter = lpin->nameIter();
	while ((name = niter++)) {
	  TLibPin tpin(d_lib.tlib(), tdir, (uint_t)lcell.totalPinCount(), testCell);
	  int idx = (*iter)++.current();
	  lcell.pin(idx).grab(tpin);
	  // process pin function and other attributes
	  if (tdir == TLD_internal) {
	    tpin.setInternalPinName(name);
	    // no attribute processing for internal pins
	    continue;
	  }
	  processPinAttr(lpin, (TLibExternalPinC*)tpin.ptr(), tdir,
			 lcell, idx,
			 outCount, pinMap, /*bundleMap=*/NULL,
			 /*busMap=*/NULL, /*busType=*/NULL, /*grpIdx=*/-1);
	}
      }
      break;
    default:
      break;
    }
  }
  if (testCell) {
    giter = testCell->groupIter();
    while ((lpin = giter++)) {
      LibAttr* busName;
      UAggType* busType = NULL;
      switch (lpin->typeID()) {
      case LIBERTY_GROUPENUM_ff:
      case LIBERTY_GROUPENUM_latch:
	processInternalSeqPins (lpin, /*isTest=*/true, intPinIter, lcell);
	break;
      case LIBERTY_GROUPENUM_bus:
	busName = lpin->findAttr("bus_type", err);
	busType = findType(busName->value()->string_val(), lcell);
	uassert (busType);
	// flow through
      case LIBERTY_GROUPENUM_pin:
	{
	  niter = lpin->nameIter();
	  while ((name = niter++)) {
	    int num = busType ? busType->size() : 1;
	    uassert (pinMap.find(name));
	    int idx = pinMap[name];
	    for (int i = 0; i < num; i++) {
	      int tidx = idx + LLoadIndexAdjust(name, NULL, &busses, busType, i);
	      if (tidx >= lcell.externalPinCount()) {
		// skip internal pins
		continue;
	      }
	      TLibPin tpin(lcell.pin(tidx));
	      // process test attributes
	      processTestPinAttr(lpin, (TLibExternalPinC*)tpin.ptr(), outCount, pinMap,
				 busType ? &busses : NULL, busType, i);
	    }
	  }
	}
	break;
      default:
	break;
      }
    }
  }
  canonicalizePinOrder(lcell, lcell.ncell().sequentialType(),
		       d_lib.bddMgr());
  // update pinMap after reordering pins
  // the lcell.pinOrder[i] stores the original index of the new i-th pin
  // we need to transform it so pinOrder[i] stores the new index of the
  // original i-th pin to make it easier to the update the index in the
  // pinMap
  int* order = new int[lcell.totalPinCount()];
#ifdef UASSERT_ENABLED
  for (
int i = 0; i < lcell.totalPinCount(); i++) {
    order[i] = -1;
  }
#endif

  int *lcellPinOrder = lcell.pinOrder();
  for (int i = 0; i < lcell.totalPinCount(); i++) {
    uassert (order[lcellPinOrder[i]] == -1);
    order[lcellPinOrder[i]] = i;
  }
  
  for (LLoadPinMapIter pmIter = pinMap.beginRandom(); pmIter; pmIter++) {
    // use the original index (pmIter.value()) to get the new
    // pin index of the corresponding pin from the pinOrder array
    pmIter.value(order[pmIter.value()]);
  }
  delete [] order;
}

void
LLoad::processModeDefinitions(const LibGroup* cellGroup, LCell& lcell)
{
  LLoadPinMap& pinMap = lcell.pinMap();
  Cudd& bddMgr = d_lib.bddMgr();
  int outCount = lcell.outPinCount();

  LibErrorT err;

  // now process the remaining details of each external pin
  LibGroupIter giter = cellGroup->groupIter();
  while (LibGroup* mode = giter++) {
    switch (mode->typeID()) {
    case LIBERTY_GROUPENUM_mode_definition :
      {
	UString name = mode->nameIter();
	UTrace modTrace(mode->filename(), mode->lineno());
	LCell::ModeVals* modeVals = lcell.addMode(name, modTrace);
	LibGroupIter miter = mode->groupIter();
	while (LibGroup* modeVal = miter++) {
	  uassert (modeVal->typeID() == LIBERTY_GROUPENUM_mode_value);
	  UString valName = modeVal->nameIter();
	  UTrace t(modeVal->filename(), modeVal->lineno());
	  LibAttr* attr = modeVal->findAttr("when", err);
	  if (!attr) {
	    LIBLOADMSGT(modeValHasNoWhenCondition, t, valName.str(), name.str());
	    continue;
	  }
	  uprintf("Processing mode definition %s - mode value %s\n", name.str(), valName.str());
	  if (!lcell.addModeValue(modeVals, valName,
				  parseFunc(attr->value()->string_val(), bddMgr, pinMap, NULL,
					    NULL, NULL, -1, outCount))) {
	    //	    LIBLOADMSGT(modeValueAlreadyExists, t, valName.str(), name.str());
	  }
	}
      }
    }
  }
}

void
LLoad::processPinDetails(const LibGroup* cellGroup, const LibGroup* testCell,
			       LCell& lcell)
{
  LLoadPinMap& pinMap = lcell.pinMap();
  LibErrorT err;
  LibGroupIter giter;
  LibGroup* lpin;
  LibNameIter niter;
  UString name;
  LibAttr* ldir;
  TLibDir tdir;

  // now process the remaining details of each external pin
  uint_t delayCount = 0;
  giter = cellGroup->groupIter();
  while ((lpin = giter++)) {
    switch (lpin->typeID()) {
    case LIBERTY_GROUPENUM_bundle:
      processBundlePinSubgroup(lpin, lcell, delayCount);
      break;
    case LIBERTY_GROUPENUM_bus:
      processBusPinSubgroup(lpin, lcell, delayCount);
      break;
    case LIBERTY_GROUPENUM_pin:
      ldir = lpin->findAttr(K_direction, err);
      tdir = (TLibDir)UString(ldir->value()->string_val()).keyword();
      niter = lpin->nameIter();
      while ((name = niter++)) {
	uassert (pinMap.find(name));
	int idx = pinMap[name];
	// process timing groups and other subgroups of pin
	// this must be done after pins have been reordered
	processPinSubgroup(lpin, idx, lcell, pinMap, /*bundles=*/NULL,
			   /*busses=*/NULL, /*busType=*/NULL,
			   /*grpIdx=*/-1, delayCount, /*dontAddIfExist=*/false);
      }
      break;
    default:
      break;
    }
  }
  lcell.ncell().delayCount(delayCount);
}

void
LLoad::createPins(const LibGroup* cellGroup, const LibGroup* coreCell, LCell& lcell)
{
  LibErrorT err;
  LibGroupIter giter;
  LibGroup* lpin;
  NDir dir;
  LibNameIter niter;
  UString name;
  LibAttr* ldir;
  TLibDir tdir;
  int* order;
  int* lcellPinOrder;
  NCell cell = lcell.ncell();

  // first create all the pins 
  giter = cellGroup->groupIter();
  while ((lpin = giter++)) {
    switch (lpin->typeID()) {
    case LIBERTY_GROUPENUM_ff:
    case LIBERTY_GROUPENUM_latch:
      addInternalPins(lpin, cell);
      break;
    case LIBERTY_GROUPENUM_bundle :
      {
	ldir = getGroupDirection(lpin);
	tdir = (TLibDir)UString(ldir->value()->string_val()).keyword();
	LibAttr* mem = lpin->findAttr("members", err);
	for (LibValIter viter = mem->valueIter(); viter; ++viter) {
	  name = viter.object()->string_val();
	  if (tdir == TLD_internal) {
	    addInternalPins(lpin, cell);
	  } else {
	    dir = (tdir == TLD_tristate) ? NDirOutput : (NDir)tdir;
	    cell.addPort(name, dir, NULL);
	  }
	}
	break;
      }
    case LIBERTY_GROUPENUM_bus :
    case LIBERTY_GROUPENUM_pin :
      {
	LibAttr* busName;
	UAggType* busType = NULL;
	if (lpin->typeID() == LIBERTY_GROUPENUM_bus) {
	  busName = lpin->findAttr("bus_type", err);
	  busType = findType(busName->value()->string_val(), lcell);
	  uassert (busType);
	}
	ldir = getGroupDirection(lpin);
	tdir = (TLibDir)UString(ldir->value()->string_val()).keyword();
	if (tdir == TLD_internal) {
	  addInternalPins(lpin, cell, busType ? busType->size() : 1);
	} else {
	  dir = (tdir == TLD_tristate) ? NDirOutput : (NDir)tdir;
	  niter = lpin->nameIter();
	  while ((name = niter++)) {
	    cell.addPort(name, dir, busType);
	  }
	}
      }
      break;
    default :
      break;
    }
  }
  if (coreCell) {
    uassert (coreCell->typeID() == LIBERTY_GROUPENUM_ff ||
	     coreCell->typeID() == LIBERTY_GROUPENUM_latch);
    addInternalPins(coreCell, cell);
  }
  
  order = new int[lcell.externalPinCount()];
#ifdef UASSERT_ENABLED
  for (int i = 0; i < lcell.externalPinCount(); i++) {
    order[i] = -1;
  }
#endif

  lcellPinOrder = lcell.pinOrder();
  for (int i = 0; i < lcell.externalPinCount(); i++) {
    uassert (order[lcellPinOrder[i]] == -1);
    order[lcellPinOrder[i]] = i;
  }
  
  cell.createPortsAndPins(order);

  // retention pin info were created the original lcell pin order.
  // fix it now.
  if (lcell.hasMVInfo()) {
    for (int i = 0; i < TRETPC_count; i++) {
      UPair<int,int> retPin = lcell.retentionPin((TRetPinClass)i);
      if (retPin.k_value() > 0) {
	NMVCell mvCell = cell.getMVCell();
	mvCell.setRetentionPin((TRetPinClass)i,
			       order[retPin.k_value()], 
			       retPin.o_value());
      }
    }
  }

  delete [] order;
  //  cell.permutePins(cell.pins(), lcell.pinOrder(), 0);

  for (UAggIter iter = lcell.allPins().nocheck().msbToLsb(); iter; iter++) {
    cell.pin(iter)->grab(lcell.pin(iter));
  }
}

void
LLoad::createPGPins(LCell& lcell)
{
  NCell cell = lcell.ncell();
  // create PG pin array here
  if (lcell.totalPGPinCount() > 0) {
    cell.createPGPinArray(lcell.outPGPinCount(), 
			  lcell.bidirPGPinCount(),
			  lcell.inPGPinCount(),
			  lcell.internalPGPinCount());
    TLibCellPG* cellPG = cell.cellPG();
    int         defP     = -1;
    int         defG     = -1;
    for (UAggIter iter = lcell.allPGPins().nocheck().msbToLsb(); iter; iter++) {
      cellPG->pgpin(iter)->grab(*lcell.pgpin(iter));

      // determine the default power and ground on the way
      if (defP == -1 && cellPG->pgpin(iter)->pgType() == TPGT_primary_power) {
	defP = iter.current();
      }
      if (defG == -1 && cellPG->pgpin(iter)->pgType() == TPGT_primary_ground) {
	defG = iter.current();
      }
    }

    NMVCell mvCell = cell.isMVCell();
    // Right now we are discarding the related_power_pin info unless a cell
    // is otherwise an MVCell already. That is, the fact that it has
    // an related_power_pin does not make it an MVCell.
    if (mvCell) {
      // copy related power pins
      for (UAggIter iter = lcell.allPins().nocheck().msbToLsb(); iter; iter++) {
	int r = lcell.relatedPower(iter.current());
	if (r >= 0) {
	  mvCell.setRelatedPower(iter.current(), r,
				 cell.logicalPinCount(), cell.cellPG()->totalPGPinCount());
	} else {
	  if (defP >= 0) {
	    mvCell.setRelatedPower(iter.current(), defP,
				   cell.logicalPinCount(), cell.cellPG()->totalPGPinCount());
	  } else {
	    uassert(0); // interesting, default power should always be there
	  }
	}
      }
      // copy related ground pins
      for (UAggIter iter = lcell.allPins().nocheck().msbToLsb(); iter; iter++) {
	int r = lcell.relatedGround(iter.current());
	if (r >= 0) {
	  mvCell.setRelatedGround(iter.current(), r,
				  cell.logicalPinCount(), cell.cellPG()->totalPGPinCount());
	} else {
	  if (defG >= 0) {
	    mvCell.setRelatedGround(iter.current(), defG,
				    cell.logicalPinCount(), cell.cellPG()->totalPGPinCount());
	  }
	}
      }

      // Here we try to infer additional properties for MV cells
      // based on the PG and related PG information.
      mvCell.inferProperty();
    }
  }
}

void
LLoad::processCellTypes(const LibGroup* lcell, LCell& tmpCell)
{
  LibGroupIter giter;
  LibGroup* group;

  giter = lcell->groupIter();
  while ((group = giter++)) {
    switch (group->typeID()) {
    case LIBERTY_GROUPENUM_type            : // ( type_name )
      {
	const UString& typeName = group->firstName();
	if (tmpCell.findType(typeName)) {
	  //	  LIBLOADMSG(multipleTypeDefsInCell, (const char*)tmpCell.ncell().name(),
	  //		     (const char*)typeName);
	} else {
	  UAggType *type = processTypeGroup(group);
	  tmpCell.quickAddType(typeName, type);
	}
      }
      break;
    default:
      break;
    }
  }
}

void
LLoad::processGeneratedClocks(const LibGroup* cellGroup, LCell& lcell)
{
  LLoadPinMap& pinMap = lcell.pinMap();
  NCell cell = lcell.ncell();
  LibGroupIter giter;
  LibGroup* group;
  char* str = NULL;
  int len = 0;

  giter = cellGroup->groupIter();
  while ((group = giter++)) {
    switch (group->typeID()) {
    case LIBERTY_GROUPENUM_generated_clock : // ( generated_clock_name )
      {
	const UString& genClkName = group->firstName();
	LibAttrIter aiter;
	LibAttr*    attr;
	const char* clkNames = NULL;
	int masterIdx = -1, divideBy = 1, multiplyBy = 1;
	bool inv = false;
	double dutyCycle = 0;
	float edge[3];
	float shift[3];
	bool hasEdges = false, hasShift = false;

	aiter = group->attrIter();
	while ((attr = aiter++)) {
	  switch (attr->typeID()) {
	  case LIBERTY_ATTRENUM_clock_pin     : // string;
	    clkNames = attr->value()->string_val();
	    break;
	  case LIBERTY_ATTRENUM_divided_by    : // virtual_attribute;
	    divideBy = (int)attr->value()->double_val();
	    break;
	  case LIBERTY_ATTRENUM_duty_cycle    : // float( > 0.0 );
	    dutyCycle = attr->value()->double_val();
	    break;
	  case LIBERTY_ATTRENUM_invert        : // boolean;
	    inv = attr->value()->bool_val();
	    break;
	  case LIBERTY_ATTRENUM_master_pin    : // string;
	    {
	      UString name = attr->value()->string_val();
	      LLoadPinMapIter pmIter = pinMap.find(name);
	      if (pmIter) {
		masterIdx = pmIter.value();
	      }
	    }
	    break;
	  case LIBERTY_ATTRENUM_multiplied_by : // virtual_attribute;
	    multiplyBy = (int)attr->value()->double_val();
	    break;
          case LIBERTY_ATTRENUM_edges :
            {
	      hasEdges = true;
              LibValIter viter = attr->valueIter();
              for (int i = 0; i < 3; ++i) {
                LibVal* v = (viter++);
                edge[i] = (float)v->int_val();
              }
            }
	    break;
          case LIBERTY_ATTRENUM_shifts :
            {
	      hasShift = true;
              LibValIter viter = attr->valueIter();
              for (int i = 0; i < 3; ++i) {
                LibVal* v = viter++;
                shift[i] = v->double_val();
              }
            }
	    break;
	  }
	}
	if (masterIdx < 0) {
	  continue;
	}
	char* clk;
	int l;
	l = strlen(clkNames) + 1;
	if (l > len) {
	  delete [] str;
	  len = l;
	  str = new char[len];
	}
	strcpy(str, clkNames);
	if (!hasShift) {
	  shift[0] = shift[1] = shift[2] = 0;
	}
	for (clk = strtok(str, ", \\\t\n"); clk; clk = strtok(NULL, ", \\\t\n")) {
	  LLoadPinMapIter pmIter = pinMap.find(clk);
	  if (!pmIter) {
	    continue;
	  }
	  int clkIdx = pmIter.value();
	  if (hasEdges) {
	    cell.addGeneratedClock(genClkName, clkIdx, masterIdx, edge, shift, inv);
	  } else {
	    cell.addGeneratedClock(genClkName, clkIdx, masterIdx, divideBy,
				   multiplyBy, shift, inv, dutyCycle);
	  }
	}
	hasEdges = false; hasShift = false;
      }
      break;
    default:
      break;
    }
  }
  delete [] str;
}

void
LLoad::instantiateCellPins(const LibGroup* cellGroup, NCell cell,
			   const LLoadRailList& railList)
{
  int output, bidir, input, internal;
  int pgOutput, pgBiDir, pgInput, pgInternal;
  LCell lcell(cell);
  LibGroup* testCell = NULL;
  LibGroup* scanCore = NULL;

  if (cell.isMVCell()) {
    lcell.makeMVInfo();
  }

  processCellTypes(cellGroup, lcell);
  countPins(cellGroup, lcell, output, bidir, input, internal, 
	    pgOutput, pgBiDir, pgInput, pgInternal,
	    testCell, scanCore);
  lcell.createPins(output, bidir, input, internal);

  if ((pgOutput + pgBiDir + pgInput + pgInternal) > 0) {
    // if pg_pin groups are defined, they take effect
    lcell.createPGPinArray(pgOutput, pgBiDir, pgInput, pgInternal);
  } else {
    // if we come here, processPins() will do nothing in creating
    // pg_pins at LCell
    if (railList.size() > 0) {
      lcell.createPGPinArray(0, 0, railList.size(), 0);
      processRailConnAttrs(lcell, cell, railList);
    }
  }
  getPinMap(cellGroup, testCell ? scanCore : NULL, lcell);
  processPGPins(cellGroup, lcell);
  processPins(cellGroup, testCell, lcell);
  createPins(cellGroup, testCell ? scanCore : NULL, lcell);
  createPGPins(lcell);
  processModeDefinitions(cellGroup, lcell);
  processPinDetails(cellGroup, testCell, lcell);
  processGeneratedClocks(cellGroup, lcell);
}

void
LLoad::preProcessCellSubgroup(const LibGroup* lcell, bool& isMvCell)
{
  LibGroupIter giter;
  LibGroup* lpin;
  LibAttr* ltype;

  giter = lcell->groupIter();
  while ((lpin = giter++)) {
    switch (lpin->typeID()) {
    case LIBERTY_GROUPENUM_pg_pin :
      {
	ltype = getGroupPGType(lpin);
	if (ltype) {
	  UString sType(ltype->value()->string_val());
	  if (sType.keyword() == K_backup_power ||
	      sType.keyword() == K_backup_ground) {
	    isMvCell = true;
	  }
	}
      }
      break;
    default :
      break;
    }
  }
}

// the following is not being used at the moment
void
LLoad::processCellSubgroup(const LibGroup* lcell, NCell cell)
{
  LibGroupIter giter;
  LibGroup* group;
  TLib* tlib = d_lib.tlib();

  // following assumes cell attributes are processeed first
  bool hasCellLeakage = cell.leakagePower(MVCRNR_INITIAL);

  giter = lcell->groupIter();
  while ((group = giter++)) {
    switch (group->typeID()) {
      // the following cell groups are supported, but handled in
      // instantiateCellPins
    case LIBERTY_GROUPENUM_bundle          : // ( bundle_name )
    case LIBERTY_GROUPENUM_bus             : // ( bus_name )
    case LIBERTY_GROUPENUM_ff              : // ( ff_name_1,ff_name_2 )
    case LIBERTY_GROUPENUM_generated_clock : // ( generated_clock_name )
    case LIBERTY_GROUPENUM_latch           : // ( latch_name_1,latch_name_2 )
    case LIBERTY_GROUPENUM_pin             : // ( pin_name_1 [,...,pin_name_N] )
    case LIBERTY_GROUPENUM_test_cell       : // ( [test_cell_name] )
    case LIBERTY_GROUPENUM_type            : // ( type_name )
      break;
    case LIBERTY_GROUPENUM_leakage_power   : // ( [leakage_power_name] )
      {
	if (!hasCellLeakage) {
	  LibAttrIter aiter = group->attrIter();
	  LibAttr* attr;
	  while ((attr = aiter++)) {
	    switch (attr->typeID()) {
	    case LIBERTY_ATTRENUM_value                         : // float( > 0.0 ) := 0.0;
	      {
		TPower oldLeak = cell.leakagePower(MVCRNR_INITIAL);
		TPower newLeak = tlib->convertLeakagePower(attr->value()->double_val());
		if (newLeak > oldLeak) {
		  cell.setLeakagePower(newLeak, MVCRNR_INITIAL);
		}
	      }
	      break;
	    default:
	      break;
	    }
	  }
	}
      }
      break;
      // the following cell groups are not supported
    case LIBERTY_GROUPENUM_ff_bank         : // ( ff_bank_name_1,ff_bank_name_2,ff_bank_name_3 )
    case LIBERTY_GROUPENUM_internal_power  : // ( internal_power_name )
    case LIBERTY_GROUPENUM_latch_bank      : // ( latch_bank_name_1,latch_bank_name_2,latch_bank_name_3 )
    case LIBERTY_GROUPENUM_memory          : // ( [memory_name] )
    case LIBERTY_GROUPENUM_mode_definition : // ( mode_definition_name )
    case LIBERTY_GROUPENUM_routing_track   : // ( routing_track_name )
    case LIBERTY_GROUPENUM_seq             : // ( seq_name_1,seq_name_2 )
    case LIBERTY_GROUPENUM_seq_bank        : // ( seq_bank_name_1,seq_bank_name_2,seq_bank_name_3 )
    case LIBERTY_GROUPENUM_state           : // ( state_name_1,state_name_2 )
    case LIBERTY_GROUPENUM_statetable      : // ( statetable_name_1,statetable_name_2 )
    default:
      if (warnNoSupport()) {
	printf("Group %s (in cell %s) on line %d of library file %s"
	       " not supported.\n", (str_t)group->type(), (str_t)cell.name(),
	       group->lineno(), group->filename());
      }
      break;
    }
  }
}
