// ****************************************************************************
//             Copyright (C) Oasys Design Systems, Inc. 2004 - 2008
//                             All rights reserved.
// ****************************************************************************
#include "UStack"
#include "Cudd"
#include "LLoadFuncParse.h"

enum FuncOperator { FO_INVALID = 0, FO_NOT, FO_XOR, FO_AND, FO_OR,
		    FO_LPAREN, FO_RPAREN };

typedef UStack<FuncOperator>   OprStack;
typedef UStack<BDD>            BddStack;

static BDD
getBdd(const char* ident, Cudd& mgr, const LLoadPinMap& idxMap, 
       LLoadBundleMap* bundles, LLoadBusMap* busses, UAggType* busType,
       int grpIdx, int idxOffset, const char* funcStr)
{
  LLoadPinMapIter iter = idxMap.find(ident);
  BDD res;
  int idx;

  if (!iter) {
    // unknown identifier found
    uprintf ("ERROR: variable %s in function expression %s not in pin list.\n",
	     ident, funcStr);
  } else {
    idx = iter.value();
    if (idx < idxOffset) {
      // variable is an output pin
      uprintf("ERROR: variable %s in function expression %s is an output pin.\n",
	      ident, funcStr);
    } else {
      idx += LLoadIndexAdjust(ident, bundles, busses, busType, grpIdx);
      res = mgr.bddVar(idx - idxOffset);
    }
  }
  return res;
}
    
static BDD
formBdd(FuncOperator opr, BDD& operand1, BDD& operand2, const char* funcStr)
{
  BDD res;

  switch (opr) {
  case FO_LPAREN:
    // error situation
    printf ("ERROR: malformed function expression %s encountered.\n", funcStr);
    break;
  case FO_NOT:
    res = !operand1;
    break;
  case FO_XOR:
    res = operand1 ^ operand2;
    break;
  case FO_AND:
    res = operand1 * operand2;
    break;
  case FO_OR:
    res = operand1 | operand2;
    break;
  default:
    // should not get here
    printf ("ERROR: unexpected operator encountered in function expression %s.\n",
	    funcStr);
    break;
  }
  return res;
}
 
static bool
implementTop(OprStack& oprS, BddStack& bddS, const char* funcStr)
{
  FuncOperator topOpr;
  BDD operand1, operand2, res;
  
  if (!oprS.top()) {
    // empty stack;
    return false;
  }
  topOpr = oprS.top();
  oprS.pop();
  if (topOpr > FO_NOT) {
    if (!bddS.top()) {
      printf ("ERROR: malformed function expression %s encountered.\n",
	      funcStr);
      return false;
    }
    operand2 = bddS.top();
    bddS.pop();
  }
  if (!bddS.top()) {
    printf ("ERROR: malformed function expression %s encountered.\n",
	    funcStr);
    return false;
  }
  operand1 = bddS.top();
  bddS.pop();
  res = formBdd(topOpr, operand1, operand2, funcStr);
  if (!res.ptr()) {
    // empty BDD; error encountered
    return false;
  } else {
    bddS.push(res);
    return true;
  }
}

static bool
pushOprOntoStack(FuncOperator opr, OprStack& oprS, BddStack& bddS,
		 const char* funcStr)
{
  FuncOperator topOpr;

  if (!oprS.top()) {    // empty stack
    oprS.push(opr);
    return true;
  }
  while (oprS.top()) {
    topOpr = oprS.top();
    if ((topOpr != FO_INVALID) && opr >= topOpr) {
      // current operator on the top of stack has a higher precedence
      // than the new operatot - carry out the operator on the top of the stack
      // stack first before pushing the new operator onto the stack 
      if (!implementTop(oprS, bddS, funcStr)) {
	return false;
      }
    } else {
      break;
    }
  }
  oprS.push(opr);
  return true;
}

static inline bool
popTillLParen(OprStack& oprS, BddStack& bddS, const char* funcStr)
{
  FuncOperator opr;

  while (oprS.top() && (opr = oprS.top()) != FO_LPAREN) {
    if (!implementTop(oprS, bddS, funcStr)) {
      return false;
    }
  }
  // must have FO_LPAREN at the top of stack; otherwise expression is wrong
  if (!oprS.top()) {
    printf ("ERROR: malformed function expression %s encountered.\n",
	    funcStr);
    return false;
  }
  oprS.pop();
  return true;
}

static inline bool
implementAll(OprStack& oprS, BddStack& bddS, const char* funcStr)
{
  while (oprS.top()) {
    if (!implementTop(oprS, bddS, funcStr)) {
      return false;
    }
  }
  return true;
}
  
BDD
parseFunc(const char* funcStr, Cudd& mgr, LLoadPinMap& idxMap, LLoadBundleMap* bundles,
	  LLoadBusMap* busses, UAggType* busType, int grpIdx, int idxOffset)
{
  uint_t len = strlen(funcStr);
  UStack<FuncOperator> oprS;
  UStack<BDD> bddS;
  char str[100];
  char *ident = str;
  short lastCharWasSpace = 0;
  short lastIdentTaken = 1;
  BDD res, tmp;

  for (uint_t i=0; i < len; i++) {
    switch (funcStr[i]) {
    case '(':
      if (ident != str) {
	// identifier exists that has not been pushed onto the stack;
	// implies an AND
	*ident = '\0';      // make sure identifier is terminated properly
	tmp = getBdd(str, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset, funcStr);
	if (!tmp.ptr()) {
	  // empty BDD; must have encountered some error
	  return res;
	}
	bddS.push(tmp);
      }
      if (!lastIdentTaken) {
	if (!pushOprOntoStack(FO_AND, oprS, bddS, funcStr)) {
	  return res;
	}
	lastIdentTaken = 1;
      }
      oprS.push(FO_LPAREN);
      ident = str;
      lastCharWasSpace = 0;
      break;
    case ')':
      if (ident != str) {
	*ident = '\0';
	tmp = getBdd(str, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset, funcStr);
	if (!tmp.ptr()) {
	  // empty BDD; must have encountered some error
	  return res;
	}
	bddS.push(tmp);
      }
      if (!popTillLParen(oprS, bddS, funcStr)) {
	return res;
      }
      ident = str;
      lastCharWasSpace = 0;
      break;
    case '!':
      if (ident != str) {
	// previous identifier not yet pushed onto the
        // stack; treat as implicit AND
	*ident = '\0';
	tmp = getBdd(str, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset, funcStr);
	if (!tmp.ptr()) {
	  // empty BDD; must have encountered some error
	  return res;
	}
	bddS.push(tmp);
      }
      if (!lastIdentTaken) {
	if (!pushOprOntoStack(FO_AND, oprS, bddS, funcStr)) {
	  return res;
	}
	lastIdentTaken = 1;
      }
      ident = str;
      lastCharWasSpace = 0;
      oprS.push(FO_NOT);
      break;
    case '\'':
      if (ident != str) {
	*ident = '\0';
	tmp = getBdd(str, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset, funcStr);
	if (!tmp.ptr()) {
	  // empty BDD; must have encountered some error
	  return res;
	}
	bddS.push(tmp);
      }
      ident = str;
      lastCharWasSpace = 0;
      oprS.push(FO_NOT);
      break;
    case '^':
      if (ident != str) {
	*ident = '\0';
	tmp = getBdd(str, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset, funcStr);
	if (!tmp.ptr()) {
	  // empty BDD; must have encountered some error
	  return res;
	}
	bddS.push(tmp);
      }
      if (!pushOprOntoStack(FO_XOR, oprS, bddS, funcStr)) {
	return res;
      }
      ident = str;
      lastIdentTaken = 1;
      lastCharWasSpace = 0;
      break;
    case '&':
    case '*':
      if (ident != str) {
	*ident = '\0';
	tmp = getBdd(str, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset, funcStr);
	if (!tmp.ptr()) {
	  // empty BDD; must have encountered some error
	  return res;
	}
	bddS.push(tmp);
      }
      if (!pushOprOntoStack(FO_AND, oprS, bddS, funcStr)) {
	return res;
      }
      ident = str;
      lastIdentTaken = 1;
      lastCharWasSpace = 0;
      break;
    case ' ':
    case '\t':
      lastCharWasSpace = 1;
      break;
    case '|':
    case '+':
      if (ident != str) {
	*ident = '\0';
	tmp = getBdd(str, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset, funcStr);
	if (!tmp.ptr()) {
	  // empty BDD; must have encountered some error
	  return res;
	}
	bddS.push(tmp);
      }
      if (!pushOprOntoStack(FO_OR, oprS, bddS, funcStr)) {
	return res;
      }
      ident = str;
      lastIdentTaken = 1;
      lastCharWasSpace = 0;
      break;
    case '0':
      if (!lastCharWasSpace && ident != str) {
	// 0 is part of an identifier
	*ident++ = '0';
	break;
      }
      if (ident != str || !bddS.empty() || !oprS.empty() ||
	  (len > i+1 && parseFunc(funcStr+i+1, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset).ptr())) {
	printf ("ERROR: constant 0 found in expression %s.\n", funcStr);
	printf ("ERROR: constants must be used by itself.\n");
	return res;
      }
      return (mgr.bddZero());
    case '1':
      if (!lastCharWasSpace && ident != str) {
	// 1 is part of an identifier
	*ident++ = '1';
	break;
      }
      if (ident != str || !bddS.empty() || !oprS.empty() ||
	  (len > i+1 && parseFunc(funcStr+i+1, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset).ptr())) {
	printf ("ERROR: constant 1 found in expression %s.\n", funcStr);
	printf ("ERROR: constants must be used by itself.\n");
	return res;
      }
      return (mgr.bddOne());
    default:
      if (ident != str && lastCharWasSpace) {
	// previous identifier has not yet been pushed onto the stack;
	// implies an AND;
	*ident = '\0';     // terminate previous identifier
	tmp = getBdd(str, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset, funcStr);
	if (!tmp.ptr()) {
	  // empty BDD; must have encountered some error
	  return res;
	}
	bddS.push(tmp);
	ident = str;
      }
      if (ident == str && !lastIdentTaken) {
	if (!pushOprOntoStack(FO_AND, oprS, bddS, funcStr)) {
	  return res;
	}
      }
      *ident++ = funcStr[i];
      lastCharWasSpace = 0;
      lastIdentTaken = 0;
      break;
    }
  }
  // push last variable onto the stack
  if (ident != str) {
    *ident = '\0';
    tmp = getBdd(str, mgr, idxMap, bundles, busses, busType, grpIdx, idxOffset, funcStr);
    if (!tmp.ptr()) {
      // empty BDD; must have encountered some error
      return res;
    }
    bddS.push(tmp);
  }
    
  if (!implementAll(oprS, bddS, funcStr)) {
    return res;
  }
  if (!bddS.empty()) {
    res = bddS.top();
    bddS.pop();
  }
  if (!bddS.empty()) {
    printf("ERROR: malformed function expression %s encountered.\n",
	   funcStr);
    return BDD();
  }
  return res;
}

